/*
 * create-flash-image
 *
 * Tetsuya OHKAWA <tetsuya@atmark-techno.com>
 *
 * Copyright (C) 2007-2008 Atmark Techno, Inc.
 */

#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <getopt.h>
#include "common.h"

#define PROG_VERSION    "1.2.0"

#define PAD (0xff)

struct file_info {
	char *name;
	unsigned long size;
};
	
static int   quiet;
static int   force;
static int   board_id = -1;
static struct file_info ifile[REGION_ID_MAX_NR];
static char *ofname;
struct region_addr addr;

static void usage(void)
{
	FILE *fp = stderr;
	int i;

	fprintf(fp, "\nusage\n");
	fprintf(fp, "  %s [options] {output_binary}\n", program_name);
	fprintf(fp, "  {output_binary}          : output filename\n");

	fprintf(fp, "options:\n");
	fprintf(fp, "  -b [%s", supported_board_names[0]);
	for (i = 1; supported_board_names[i] != NULL; i++) 
		fprintf(fp, "|%s", supported_board_names[i]);
	fprintf(fp, "]\n                     : board type\n");
	fprintf(fp, "  --fpga={binary}       : binary filename of fpga region\n");
	fprintf(fp, "  --ipl={binary}        : binary filename of ipl region\n");
	fprintf(fp, "  --bootloader={binary} : binary filename of bootloader region\n");
	fprintf(fp, "  --image={binary}      : binary filename of image region\n");
	fprintf(fp, "  --kernel={binary}     : binary filename of kernel region\n");
	fprintf(fp, "  --userland={binary}   : binary filename of userland region\n");
	fprintf(fp, "  --recovery={binary}   : binary filename of recovery region\n");
	fprintf(fp, "  --free1={binary}      : binary filename of free1 region\n");
	fprintf(fp, "  --free2={binary}      : binary filename of free2 region\n");
	fprintf(fp, "  -q, --quiet           : print nothing, only return an exit status\n");
	fprintf(fp, "  -f, --force           : overwrite existing output file\n");
	fprintf(fp, "  -v, --version         : show version\n");
	fprintf(fp, "  -h, --help            : show help\n");
}

static void version(void)
{
	fprintf(stderr, "%s version %s\n", program_name, PROG_VERSION);
}

static int parse_option(int argc, char *argv[])
{
	char *s;
	int i;
	int opt;
	char *board_name = NULL;
	static struct option longopts[] = {{"", 1, NULL, 'b'},
					   {"fpga", 1, NULL, REGION_ID_FPGA},
					   {"ipl", 1, NULL, REGION_ID_IPL},
					   {"bootloader", 1, NULL, REGION_ID_BOOTLOADER},
					   {"kernel", 1, NULL, REGION_ID_KERNEL},
					   {"userland", 1, NULL, REGION_ID_USERLAND},
					   {"recovery", 1, NULL, REGION_ID_RECOVERY},
					   {"license", 1, NULL, REGION_ID_LICENSE},
					   {"image", 1, NULL, REGION_ID_IMAGE},
					   {"config", 1, NULL, REGION_ID_CONFIG},
					   {"free1", 1, NULL, REGION_ID_FREE1},
					   {"free2", 1, NULL, REGION_ID_FREE2},
					   {"force", 0, NULL, 'f'},
					   {"quiet", 0, NULL, 'q'}, 
					   {"help", 0, NULL, 'h'}, 
					   {"version", 0, NULL, 'v'}, 
					   {0, 0, 0, 0}};

	while ((opt = getopt_long(argc, argv, "1:2:3:4:5:6:7:8:9:b:fqhv", longopts, NULL)) != -1) {
		switch (opt) {
		case REGION_ID_FPGA:
			ifile[REGION_ID_FPGA].name = optarg;
			break;
		case REGION_ID_IPL:
			ifile[REGION_ID_IPL].name = optarg;
			break;
		case REGION_ID_BOOTLOADER:
			ifile[REGION_ID_BOOTLOADER].name = optarg;
			break;
		case REGION_ID_KERNEL:
			ifile[REGION_ID_KERNEL].name = optarg;
			break;
		case REGION_ID_USERLAND:
			ifile[REGION_ID_USERLAND].name = optarg;
			break;
		case REGION_ID_RECOVERY:
			ifile[REGION_ID_RECOVERY].name = optarg;
			break;
		case REGION_ID_IMAGE:
			ifile[REGION_ID_IMAGE].name = optarg;
			break;
		case REGION_ID_FREE1:
			ifile[REGION_ID_FREE1].name = optarg;
			break;
		case REGION_ID_FREE2:
			ifile[REGION_ID_FREE2].name = optarg;
			break;
		case 'b':
			board_name = optarg;
			break;
		case 'f':
			force = 1;
			break;
		case 'q':
			quiet = 1;
			break;
		case 'v':
			version();
			exit(0);
		case 'h':
			usage();
			exit(0);
		case '?':
			LOG_ERROR("unknown option: %c\n", optopt);
			return 1;
		}
	}

	for ( ; optind < argc; optind++) {
		if (!ofname) {
			ofname = argv[optind];
		} else {
			LOG_ERROR("too many argument: %s\n", argv[optind]);
			return 1;
		}
	}
	if (!ofname) {
		LOG_ERROR("no output filename\n");
		return 1;
	}
	if (!force) {
		if (access(ofname, F_OK) == 0) {
			LOG_ERROR("the output file already exists");
			return 1;
		}
	}

	if (board_name != NULL) {
		for (i = 0; (s = supported_board_names[i]) != NULL; i++) {
			if (strcmp(s, board_name) == 0) {
				board_id = i;
				break;
			}
		}
	}
	if (board_id == -1) {
		LOG_ERROR("missing board type\n");
		return 1;
	}
	
	return 0;
}

static int name_to_id(const char *name)
{
	int i;

	for (i = 0; i < REGION_ID_MAX_NR; i++)
		if (strcmp(name, region_pairs[i].name) == 0)
			return region_pairs[i].id;
	return -1;
}

static int open_file(char *name,  int flag)
{
	int fd, mode;

	if (flag & O_CREAT) {
		mode = S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH;
		fd = open(name, flag, mode);
	} else {
		fd = open(name, flag);
	}
	if (fd == -1)
		LOG_ERROR("cannot open %s\n", name);

	return fd;
}

static off_t get_file_size(int region)
{
	int fd;
	off_t size;

	if (ifile[region].name == NULL)
		return 0;

	fd = open_file(ifile[region].name, O_RDONLY);
	if (fd < 0)
		return -1;

	size = lseek(fd, 0, SEEK_END);
	close(fd);

	return size;
}

static int check_region_size(void)
{
	int i;
	int id = 0;

	for (i = 0; flash_map[board_id][i].name != NULL; i++) {
		id = name_to_id(flash_map[board_id][i].name);
		if (ifile[id].name == NULL)
			continue;

		ifile[id].size = get_file_size(id);
		if ((off_t)ifile[id].size == -1)
			return 1;
		
		if (ifile[id].size > flash_map[board_id][i].addr.size) {
			LOG_ERROR("binary size is over %s-region (%s)\n",
				  flash_map[board_id][i].name, ifile[id].name);
			return 1;
		}
	}

	return 0;
}

static int out_region_binary(int region, void *omap)
{
	int fd;
	int id;
	void *imap, *p = omap;

	id = name_to_id(flash_map[board_id][region].name);
	if (ifile[id].name != NULL) {
		if ((fd = open_file(ifile[id].name, O_RDONLY)) < 0)
			return 1;
	
		imap = mmap(NULL, ifile[id].size, PROT_READ, MAP_SHARED, fd, 0);
		if (imap == MAP_FAILED) {
			LOG_ERROR("mmap failed: %s\n", strerror(errno));
			close(fd);
			return 1;
		}
		close(fd);

		p = mempcpy(omap, imap, ifile[id].size);

		munmap(imap, ifile[id].size);

		LOG_INFO("%s (%s, 0x%08lX - 0x%08lX) combined (%s)\n",
			 supported_board_names[board_id],
			 flash_map[board_id][region].name,
			 flash_map[board_id][region].addr.start,
			 flash_map[board_id][region].addr.start +
			 flash_map[board_id][region].addr.size - 1,
			 ifile[id].name);
	}

	memset(p, PAD, flash_map[board_id][region].addr.size - ifile[id].size);

	return 0;
}

static size_t get_flash_size(void)
{
	int i;

	for (i = 0; flash_map[board_id][i].name != NULL; i++)
		if(!strcmp(flash_map[board_id][i].name, REGION_NAME_ALL))
			return flash_map[board_id][i].addr.size;

	return 0;
}

static int merge_binary(void)
{
	int i;
	int fd;
	void *map;
	int ret;
	size_t flash_size;
	
	if ((fd = open_file(ofname, O_RDWR|O_CREAT|O_TRUNC)) < 0)
		return 1;

	flash_size = get_flash_size();

	/* create a file with size we want */
	lseek(fd, flash_size - 1, SEEK_SET);
	ret = write(fd, "", 1);
	if (ret != 1)
		return 1;
	lseek(fd, 0, SEEK_SET);

	map = mmap(NULL, flash_size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
	if (map == MAP_FAILED) {
		LOG_ERROR("mmap failed: %s\n", strerror(errno));
		close(fd);
		return 1;
	}
	close(fd);

	/* merge */
	for (i = 0; flash_map[board_id][i].name != NULL; i++) {
		if ((strcmp(flash_map[board_id][i].name, REGION_NAME_ALL) == 0) ||
			(strcmp(flash_map[board_id][i].name, REGION_NAME_ALL_WITHOUT_CONFIG) == 0))
			continue;
		ret = out_region_binary(i, map + flash_map[board_id][i].addr.start);
		if (ret) {
			munmap(map, flash_size);
			return 1;
		}
	}
	munmap(map, flash_size);

	return 0;
}

int main(int argc, char *argv[])
{

	program_name = argv[0];

	if (parse_option(argc, argv))
		return 1;

  	if (check_region_size())
		return 1;

	if (merge_binary()) {
		LOG_ERROR("failure\n");
		return 1;
	}

	return 0;
}
